'* xPL VIOM
'*
'* Copyright (C) 2004 John Bent
'* http://www.xpl.myby.co.uk/
'*
'* This program is free software; you can redistribute it and/or
'* modify it under the terms of the GNU General Public License
'* as published by the Free Software Foundation; either version 2
'* of the License, or (at your option) any later version.
'* 
'* This program is distributed in the hope that it will be useful,
'* but WITHOUT ANY WARRANTY; without even the implied warranty of
'* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'* GNU General Public License for more details.
'*
'* You should have received a copy of the GNU General Public License
'* along with this program; if not, write to the Free Software
'* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
Option Strict On

#Const VIOM_DEBUG = 0

Public Class viom
  Inherits CommBase

  Private Const MAX_BUFFER_SIZE As Integer = 255
  Private Const ERROR_PREFIX As String = "-"
  Private Const INPUT_PREFIX As String = "+"

  Private Buffer(MAX_BUFFER_SIZE) As Byte
  Private CurrentBufferSize As Integer
  Public EventLog As EventLog
  Private WithEvents myXplListener As XplListener

  ' String to hold current state of all inputs
  Private InputStates As String


  Public Sub Startup(ByVal instanceNumber As Integer)
    Log("xPLVIOM is starting.")
    InputStates = "0000000000000000"
    myXplListener = New XplListener("johnb-viom", instanceNumber, EventLog)
    If myXplListener.AwaitingConfiguration Then
      myXplListener.ConfigItems.Add("port", "1", xpllib.xplConfigTypes.xReconf)
      myXplListener.ConfigItems.Add("host", "", xpllib.xplConfigTypes.xOption)
    End If

    myXplListener.Listen()
    If Not myXplListener.AwaitingConfiguration Then
      Initialise()
    Else
      Log("Waiting to be configured.")
    End If
  End Sub

  Public Sub Shutdown()
    Log("xPLVIOM is shutting down.")
    MyBase.Close()
    myXplListener.Dispose()
    myXplListener = Nothing
  End Sub

  Private Sub Initialise()
    Log("Initialising VIOM...")
    Try
      MyBase.Close()
    Catch ex As Exception
    End Try
    Try
      If Not MyBase.Open Then
        Log("Failed to open COM port. See event log for details.")
        Throw New Exception("Failed to open COM" & myXplListener.ConfigItems.Item("comport"))
      End If
      CurrentBufferSize = 0
      ' Set VIOM to short message mode
      Send(Encoding.ASCII.GetBytes("CMS" & vbCrLf))
      ' Set the prefix for input changes to the + character
      Send(Encoding.ASCII.GetBytes("CIP1" & INPUT_PREFIX & vbCrLf))
      ' Set the error prefix
      Send(Encoding.ASCII.GetBytes("CEA" & ERROR_PREFIX & vbCrLf))
      ' Enable all outputs for control via RS232
      For Counter As Integer = 1 To 16
        Send(Encoding.ASCII.GetBytes("WO" & Counter.ToString & "N3" & vbCrLf))
      Next
      ' Enable input change reporting
      Send(Encoding.ASCII.GetBytes("CIC1" & vbCrLf))
    Catch ex As Exception
      EventLog.WriteEntry("Error initialising xPL VIOM: " & ex.tostring, EventLogEntryType.Error)
    End Try
    Log("Initialisation completed.")
  End Sub

  Private Sub myXplListener_XplConfigDone() Handles myXplListener.XplConfigDone
    Log("Received xPL config.response.")
    Initialise()
  End Sub

  Protected Overrides Function CommSettings() As CommBaseSettings
    Dim cs As New CommBaseSettings
    cs.SetStandard("COM" & myXplListener.ConfigItems.Item("comport") & ":", 9600, CommBase.Handshake.none)
    Log("Serial port is " & cs.port)
    Return cs
  End Function

  Protected Overrides Sub OnRxChar(ByVal c As Byte)
    Log("Received byte: " & CInt(c))
    ' Add the received character to the buffer
    If CurrentBufferSize < MAX_BUFFER_SIZE Then
      Buffer(CurrentBufferSize) = c
      CurrentBufferSize += 1

      ' Look for a CrLf
      If CurrentBufferSize > 1 Then
        If Buffer(CurrentBufferSize - 1) = 10 And Buffer(CurrentBufferSize - 2) = 13 Then
          ProcessLine()
        End If
      End If
    End If
  End Sub

  Private Sub ProcessLine()
    ' Called whenever the buffer has a complete line of data for us to read
    Log("Processing line: " & Encoding.ASCII.GetString(Buffer, 0, CurrentBufferSize))
    ' Look at the first character and determine what to do
    Select Case Buffer(0)
      Case 43
        Dim newstates As String = Encoding.ASCII.GetString(Buffer)
        newstates = newstates.Substring(1, 16)
        ' See if any inputs have changed
        For Counter As Integer = 0 To 15
          If newstates.Chars(Counter) <> InputStates.Chars(Counter) Then
            myXplListener.SendMessage("xpl-trig", "*", "sensor.basic", "device=" & (Counter + 1).ToString & vbLf & "type=input" & vbLf & "current=" & GetState(newstates.Chars(Counter)) & vbLf)
          End If
        Next
        InputStates = newstates
    End Select
    ' Clear the buffer
    For Counter As Integer = 0 To MAX_BUFFER_SIZE - 1
      Buffer(counter) = 0
    Next
    CurrentBufferSize = 0
  End Sub

  Private Sub myXplListener_XplMessageReceived(ByVal sender As Object, ByVal e As xpllib.XplListener.XplEventArgs) Handles myXplListener.XplMessageReceived
    Dim x As xpllib.XplMsg = e.XplMsg
    ' Look for command messages
    Select Case x.XPL_Msg(0).Section.ToLower
      Case "xpl-cmnd"
        Select Case x.Schema.msgClass
          Case "control"
            ProcessControlMessage(x)
        End Select
    End Select

  End Sub

  Private Sub ProcessControlMessage(ByVal x As XplMsg)
    Log("Processing received control.basic message...")
    Dim d As String = x.GetParam(1, "device")
    Dim t As String = x.GetParam(1, "type")
    Dim c As String = x.GetParam(1, "current")
    Select Case t.ToLower
      Case "input"

      Case "output"
        Select Case c.ToLower
          Case "on", "high"
            If d = "all" Then
              Send(Encoding.ASCII.GetBytes("ZA1111111111111111" & vbCrLf))
            Else
              Send(Encoding.ASCII.GetBytes("XA" & d & vbCrLf))
            End If
          Case "off", "low"
            If d = "all" Then
              Send(Encoding.ASCII.GetBytes("ZA0000000000000000" & vbCrLf))
            Else
              Send(Encoding.ASCII.GetBytes("XB" & d & vbCrLf))
            End If
          Case "request"

        End Select
    End Select
  End Sub

  Private Function GetState(ByVal state As Char) As String
    Select Case state.ToString
      Case "1"
        Return "high"
      Case "0"
        Return "low"
    End Select
  End Function

  Private Sub Log(ByVal str As String)
#If VIOM_DEBUG Then
    Dim logFile As TextWriter = File.AppendText("c:\viom-debug.log")
    logFile.WriteLine(Now.ToString("dd/MMM/yyyy HH:mm:ss") & " " & str)
    logFile.Close()
#End If
  End Sub

End Class
